# coding: utf-8
"""Extension properties for Model, Room, Face, Shade, Aperture, Door.

These objects hold all attributes assigned by extensions like honeybee-radiance
and honeybee-energy.  Note that these Property objects are not intended to exist
on their own but should have a host object.
"""


class _Properties(object):
    """Base class for all Properties classes.

    Args:
        host: A honeybee-core geometry object that hosts these properties
            (ie. Model, Room, Face, Shade, Aperture, Door).
    """
    _exclude = set(
        ('host', 'move', 'rotate', 'rotate_xy', 'reflect', 'scale', 'is_equivalent',
         'add_prefix', 'reset_to_default', 'to_dict', 'apply_properties_from_dict',
         'ToString'))

    def __init__(self, host):
        """Initialize properties."""
        self._host = host

    @property
    def host(self):
        """Get the object hosting these properties."""
        return self._host

    @property
    def _extension_attributes(self):
        return (atr for atr in dir(self) if not atr.startswith('_')
                and atr not in self._exclude)

    def move(self, moving_vec):
        """Apply a move transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be moved alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        moved if the host aperture is moved.

        Args:
            moving_vec: A ladybug_geometry Vector3D with the direction and distance
                to move the face.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'move'):
                continue
            try:
                var.move(moving_vec)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to move {}: {}'.format(var, e))

    def rotate(self, axis, angle, origin):
        """Apply a rotation transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be rotated alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        rotated if the host aperture is rotated.

        Args:
            axis: A ladybug_geometry Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'rotate'):
                continue
            try:
                var.rotate(axis, angle, origin)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to rotate {}: {}'.format(var, e))

    def rotate_xy(self, angle, origin):
        """Apply a rotation in the XY plane to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be rotated alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        rotated if the host aperture is rotated.

        Args:
            angle: An angle in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'rotate_xy'):
                continue
            try:
                var.rotate_xy(angle, origin)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to rotate {}: {}'.format(var, e))

    def reflect(self, plane):
        """Apply a reflection transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be reflected alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        reflected if the host aperture is reflected.

        Args:
            plane: A ladybug_geometry Plane across which the object will
                be reflected.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'reflect'):
                continue
            try:
                var.reflect(plane)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to reflect {}: {}'.format(var, e))

    def scale(self, factor, origin=None):
        """Apply a scale transform to extension attributes.

        This is useful in cases where extension attributes possess geometric data
        that should be scaled alongside the host object. For example, dynamic
        geometry within the honeybee-radiance state of an aperture should be
        scaled if the host aperture is scaled.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point3D representing the origin from which
                to scale. If None, it will be scaled from the World origin (0, 0, 0).
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'scale'):
                continue
            try:
                var.scale(factor, origin)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to scale {}: {}'.format(var, e))

    def is_equivalent(self, other_properties):
        """Get a dictionary noting the equivalency of these properties to other ones.

        The keys of this dictionary will note the name of each extension (eg.
        energy, radiance) and the values will be a boolean for whether the
        extension properties are equivalent or not.

        Args:
            other_properties: Properties of another object for which equivalency
                will be tested.
        """
        eq_dict = {}
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'is_equivalent'):
                continue
            other_var = getattr(other_properties, atr)
            try:
                eq_dict[atr] = var.is_equivalent(other_var)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed test is_equivalent for {}: {}'.format(var, e))
        return eq_dict

    def _update_by_sync(self, change, existing_prop, updated_prop):
        """Update properties using change instructions and existing/updated objects."""
        for atr in self._extension_attributes:
            up_atr = 'update_{}'.format(atr)
            if up_atr in change:
                var = getattr(updated_prop, atr) if change[up_atr] \
                    else getattr(existing_prop, atr)
                if not hasattr(var, 'duplicate'):
                    continue
                try:
                    setattr(self, '_' + atr, var.duplicate(self.host))
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    raise Exception('Failed to duplicate {}: {}'.format(var, e))

    def _duplicate_extension_attr(self, original_properties):
        """Duplicate the attributes added by extensions.

        This method should be called within the duplicate or __copy__ methods of
        each honeybee-core geometry object after the core object has been duplicated.
        This method only needs to be called on the new (duplicated) core object and
        the extension properties of the original core object should be passed to
        this method as the original_properties.

        Args:
            original_properties: The properties object of the original core
                object from which the duplicate was derived.
        """
        for atr in self._extension_attributes:
            var = getattr(original_properties, atr)
            if not hasattr(var, 'duplicate'):
                continue
            try:
                setattr(self, '_' + atr, var.duplicate(self.host))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to duplicate {}: {}'.format(var, e))

    def _add_prefix_extension_attr(self, prefix):
        """Change the name extension attributes unique to this object by adding a prefix.

        This is particularly useful in workflows where you duplicate and edit
        a starting object and then want to combine it with the original object
        into one Model (like making a model of repeated rooms).

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the object and does not add the prefix to attributes that are
        shared across several objects.

        Args:
            prefix: Text that will be inserted at the start of the extension attributes'
                name. It is recommended that this name be short to avoid maxing
                out the 100 allowable characters for honeybee names.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'add_prefix'):
                continue
            try:
                var.add_prefix(prefix)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to add prefix to {}: {}'.format(var, e))

    def _reset_extension_attr_to_default(self):
        """Reset all extension attributes for the object to be default.

        This is useful in cases where properties that are hard-assigned to a specific
        object might be illegal in combination with other properties and so they
        should be reset upon the setting of the other properties. For example,
        setting a Face type to AirBoundary typically makes certain types of energy
        and radiance properties illegal. So calling this function whenever setting
        Face type to AirBoundary will reset the extension attributes to the legal
        default values.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'reset_to_default'):
                continue
            try:
                var.reset_to_default()
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to reset_to_default for {}: {}'.format(var, e))

    def _add_extension_attr_to_dict(self, base, abridged, include):
        """Add attributes for extensions to the base dictionary.

        This method should be called within the to_dict method of each honeybee-core
        geometry object.

        Args:
            base: The dictionary of the core object without any extension attributes.
                This method will add extension attributes to this dictionary. For
                example, energy properties will appear under base['properties']['energy'].
            abridged: Boolean to note whether the attributes of the extensions should
                be abridged (True) or full (False). For example, if a Room's energy
                properties are abridged, the program_type attribute under the energy
                properties dictionary will just be the identifier of the program_type. If
                it is full (not abridged), the program_type will be a complete
                dictionary following the ProgramType schema. Abridged dictionaries
                should be used within the Model.to_dict but full dictionaries should
                be used within the to_dict methods of individual objects.
            include: List of properties to filter keys that must be included in
                output dictionary. For example ['energy'] will include 'energy' key if
                available in properties to_dict. By default all the keys will be
                included. To exclude all the keys from extensions use an empty list.
        """
        attr = include if include is not None else self._extension_attributes
        for atr in attr:
            var = getattr(self, atr)
            if not hasattr(var, 'to_dict'):
                continue
            try:
                base.update(var.to_dict(abridged))
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to convert {} to a dict: {}'.format(var, e))
        return base

    def _load_extension_attr_from_dict(self, property_dict):
        """Get attributes for extensions from a dictionary of the properties.

        This method should be called within the from_dict method of each honeybee-core
        geometry object. Specifically, this method should be called on the core
        object after it has been created from a dictionary but lacks any of the
        extension attributes in the dictionary.

        Args:
            property_dict: A dictionary of properties for the object (ie.
                FaceProperties, RoomProperties). These will be used to load
                attributes from the dictionary and assign them to the object on
                which this method is called.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if not hasattr(var, 'from_dict'):
                continue
            try:
                setattr(self, '_' + atr, var.__class__.from_dict(
                    property_dict[atr], self.host))
            except KeyError:
                pass  # the property_dict possesses no properties for that extension

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        """Properties representation."""
        return 'BaseProperties'


class ModelProperties(_Properties):
    """Honeybee Model Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        model = Model('New Elementary School', list_of_rooms)
        model.properties -> ModelProperties
        model.properties.radiance -> ModelRadianceProperties
        model.properties.energy -> ModelEnergyProperties
    """

    def to_dict(self, include=None):
        """Convert properties to dictionary.

        Args:
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ModelProperties'}
        attr = include if include is not None else self._extension_attributes
        for atr in attr:
            var = getattr(self, atr)
            if not hasattr(var, 'to_dict'):
                continue
            try:
                base.update(var.to_dict())  # no abridged dictionary for model
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception('Failed to convert {} to a dict: {}'.format(var, e))
        return base

    def apply_properties_from_dict(self, data):
        """Apply extension properties from a Model dictionary to the host Model.

        Args:
            data: A dictionary representation of an entire honeybee-core Model.
        """
        for atr in self._extension_attributes:
            if atr not in data['properties'] or data['properties'][atr] is None:
                continue
            var = getattr(self, atr)
            if var and not hasattr(var, 'apply_properties_from_dict'):
                continue
            try:
                var.apply_properties_from_dict(data)
            except Exception as e:
                import traceback
                traceback.print_exc()
                raise Exception(
                    'Failed to apply {} properties to the Model: {}'.format(atr, e))

    def _check_for_extension(self, extension_name, detailed=False):
        """Check the validity of the model for a specific extension.

        Args:
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).
            extension_name: Text for the name of the extension to be checked.
                This value should always be lowercase to match the name of the
                extension package. Some common honeybee extension names that can
                be input here if they are installed include:

                * radiance
                * energy
                * doe2
                * ies
                * idaice
        """
        msgs = []
        for atr in self._extension_attributes:
            if extension_name == atr:
                check_msg = None
                try:
                    var = getattr(self, atr)
                except AttributeError as e:
                    raise ImportError(
                        'Extension for {} is not installed or has not been set up '
                        'for model validation.\n{}'.format(var, e))
                if not hasattr(var, 'check_for_extension'):
                    raise NotImplementedError(
                        'Extension for {} does not have validation routines.'.format(var))
                try:
                    check_msg = var.check_for_extension(
                        raise_exception=False, detailed=detailed)
                    if detailed and check_msg is not None:
                        msgs.append(check_msg)
                    elif check_msg != '':
                        f_msg = 'Attributes for {} are invalid.\n{}'.format(atr, check_msg)
                        msgs.append(f_msg)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    raise Exception('Failed to check_for_extension '
                                    'for {}: {}'.format(var, e))
        return msgs

    def _check_all_extension_attr(self, detailed=False, all_ext_checks=False):
        """Check the attributes of all extensions.

        This method should be called within the check_all method of the Model object
        to ensure that the check_all functions of any extension model properties
        are also called.

        Args:
            detailed: Boolean for whether the returned object is a detailed list of
                dicts with error info or a string with a message. (Default: False).
        """
        msgs = []
        for atr in self._extension_attributes:
            # get the extension attributes
            check_msg = None
            var = getattr(self, atr)
            # use the check_generic function if it is available
            if not all_ext_checks and hasattr(var, 'check_generic'):
                try:
                    check_msg = var.check_generic(
                        raise_exception=False, detailed=detailed)
                    if detailed and check_msg is not None:
                        msgs.append(check_msg)
                    elif check_msg != '':
                        f_msg = \
                            'Attributes for {} are invalid.\n{}'.format(atr, check_msg)
                        msgs.append(f_msg)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    raise Exception('Failed to check_generic for {}: {}'.format(var, e))
            elif hasattr(var, 'check_all'):  # use the check_all function
                try:
                    try:
                        check_msg = var.check_all(
                            raise_exception=False, detailed=detailed)
                    except TypeError:  # no option available for detailed error message
                        check_msg = var.check_all(raise_exception=False)
                    if detailed and check_msg is not None:
                        msgs.append(check_msg)
                    elif check_msg != '':
                        f_msg = \
                            'Attributes for {} are invalid.\n{}'.format(atr, check_msg)
                        msgs.append(f_msg)
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    raise Exception('Failed to check_all for {}: {}'.format(var, e))
        return msgs

    def _check_func_from_code(self, error_code):
        """Get a check function for a specific error code that exists in an extension.

        Args:
            error_code: Text for the error code for which the check will be performed.
                This can be the value under the "code" key in the dictionary of
                the validation error whenever the detailed option is used.
        """
        for atr in self._extension_attributes:
            var = getattr(self, atr)
            if hasattr(var, 'ERROR_MAP'):
                err_map = getattr(var, 'ERROR_MAP')
                try:
                    check_func = err_map[error_code]
                    check_func = getattr(var, check_func)
                    return check_func
                except KeyError:  # not the correct extension for the code
                    pass

    def __repr__(self):
        """Properties representation."""
        return 'ModelProperties: {}'.format(self.host.display_name)


class RoomProperties(_Properties):
    """Honeybee Room Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        room = Room('Bedroom', geometry)
        room.properties -> RoomProperties
        room.properties.radiance -> RoomRadianceProperties
        room.properties.energy -> RoomEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'RoomProperties'} if not abridged else \
            {'type': 'RoomPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Room (eg. single-room HVAC systems) and does not add the
        prefix to attributes that are shared across several Rooms (eg. ConstructionSets).

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned at the level of this Room to default.

        This typically means erasing any ConstructionSets or ModifierSets assigned
        to this Room among other properties.
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'RoomProperties: {}'.format(self.host.display_name)


class FaceProperties(_Properties):
    """Honeybee Face Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        face = Face('South Bedroom Wall', geometry)
        face.properties -> FaceProperties
        face.properties.radiance -> FaceRadianceProperties
        face.properties.energy -> FaceEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary besides Face type
                and boundary_condition. If None all the available keys will be included.
        """
        base = {'type': 'FaceProperties'} if not abridged else \
            {'type': 'FacePropertiesAbridged'}
        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Face and does not add the prefix to attributes that are
        shared across several Faces.

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned at the level of this Face to default.

        This typically means erasing any Constructions or Modifiers assigned to this
        Face (having them instead assigned by ConstructionSets and ModifierSets).
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'FaceProperties: {}'.format(self.host.display_name)


class ApertureProperties(_Properties):
    """Honeybee Aperture Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        aperture = Aperture('Window to My Soul', geometry)
        aperture.properties -> ApertureProperties
        aperture.properties.radiance -> ApertureRadianceProperties
        aperture.properties.energy -> ApertureEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ApertureProperties'} if not abridged else \
            {'type': 'AperturePropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Aperture and does not add the prefix to attributes that are
        shared across several Apertures.

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned to this Aperture to default.

        This typically means erasing any Constructions or Modifiers assigned to this
        Aperture (having them instead assigned by ConstructionSets and ModifierSets).
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'ApertureProperties: {}'.format(self.host.display_name)


class DoorProperties(_Properties):
    """Honeybee Door Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        door = Door('Front Door', geometry)
        door.properties -> DoorProperties
        door.properties.radiance -> DoorRadianceProperties
        door.properties.energy -> DoorEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'DoorProperties'} if not abridged else \
            {'type': 'DoorPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Door and does not add the prefix to attributes that are
        shared across several Doors.

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned to this Door to default.

        This typically means erasing any Constructions or Modifiers assigned to this
        Door (having them instead assigned by ConstructionSets and ModifierSets).
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'DoorProperties: {}'.format(self.host.display_name)


class ShadeProperties(_Properties):
    """Honeybee Shade Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        shade = Shade('Deep Overhang', geometry)
        shade.properties -> ShadeProperties
        shade.properties.radiance -> ShadeRadianceProperties
        shade.properties.energy -> ShadeEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ShadeProperties'} if not abridged else \
            {'type': 'ShadePropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the Shade and does not add the prefix to attributes that are
        shared across several Shades.

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned at the level of this Shade to default.

        This typically means erasing any Constructions or Modifiers assigned to this
        Shade (having them instead assigned by ConstructionSets and ModifierSets).
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'ShadeProperties: {}'.format(self.host.display_name)


class ShadeMeshProperties(_Properties):
    """Honeybee ShadeMesh Properties.

    This class will be extended by extensions.

    Usage:

    .. code-block:: python

        shade = ShadeMesh('Hilly Terrain', geometry)
        shade.properties -> ShadeMeshProperties
        shade.properties.radiance -> ShadeMeshRadianceProperties
        shade.properties.energy -> ShadeMeshEnergyProperties
    """

    def to_dict(self, abridged=False, include=None):
        """Convert properties to dictionary.

        Args:
            abridged: Boolean to note whether the full dictionary describing the
                object should be returned (False) or just an abridged version (True).
                Default: False.
            include: A list of keys to be included in dictionary.
                If None all the available keys will be included.
        """
        base = {'type': 'ShadeMeshProperties'} if not abridged else \
            {'type': 'ShadeMeshPropertiesAbridged'}

        base = self._add_extension_attr_to_dict(base, abridged, include)
        return base

    def add_prefix(self, prefix):
        """Change the identifier extension attributes unique to this object by adding a prefix.

        Notably, this method only adds the prefix to extension attributes that must
        be unique to the ShadeMesh and does not add the prefix to attributes that are
        shared across several ShadeMeshes (eg. modifier).

        Args:
            prefix: Text that will be inserted at the start of extension attribute identifiers.
        """
        self._add_prefix_extension_attr(prefix)

    def reset_to_default(self):
        """Reset the extension properties assigned to this ShadeMesh to default.

        This typically means erasing any Constructions or Modifiers assigned to this
        ShadeMesh.
        """
        self._reset_extension_attr_to_default()

    def __repr__(self):
        """Properties representation."""
        return 'ShadeMeshProperties: {}'.format(self.host.display_name)



"""Boundary Condition for Face, Aperture, Door."""
import re

from .typing import float_in_range, tuple_with_length
from .altnumber import autocalculate


class _BoundaryCondition(object):
    """Base boundary condition class."""

    __slots__ = ()

    def __init__(self):
        """Initialize Boundary condition."""

    @property
    def name(self):
        """Get the name of the boundary condition (ie. 'Outdoors', 'Ground')."""
        return self.__class__.__name__

    @property
    def view_factor(self):
        """Get the view factor to the ground."""
        return 'autocalculate'

    @property
    def sun_exposure_idf(self):
        """Get a text string for sun exposure, which is write-able into an IDF."""
        return 'NoSun'

    @property
    def wind_exposure_idf(self):
        """ Get a text string for wind exposure, which is write-able into an IDF."""
        return 'NoWind'

    def to_dict(self):
        """Get the boundary condition as a dictionary."""
        return {'type': self.name}

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        return self.name


class Outdoors(_BoundaryCondition):
    """Outdoor boundary condition.

    Args:
        sun_exposure: A boolean noting whether the boundary is exposed to sun.
            Default: True.
        wind_exposure: A boolean noting whether the boundary is exposed to wind.
            Default: True.
        view_factor: A number between 0 and 1 for the view factor to the ground.
            This input can also be an Autocalculate object to signify that the view
            factor automatically calculated.  Default: autocalculate.
    """

    __slots__ = ('_sun_exposure', '_wind_exposure', '_view_factor')

    def __init__(self, sun_exposure=True, wind_exposure=True,
                 view_factor=autocalculate):
        """Initialize Outdoors boundary condition."""
        assert isinstance(sun_exposure, bool), \
            'Input sun_exposure must be a Boolean. Got {}.'.format(type(sun_exposure))
        self._sun_exposure = sun_exposure
        assert isinstance(wind_exposure, bool), \
            'Input wind_exposure must be a Boolean. Got {}.'.format(type(wind_exposure))
        self._wind_exposure = wind_exposure
        if view_factor == autocalculate:
            self._view_factor = autocalculate
        else:
            self._view_factor = float_in_range(
                view_factor, 0.0, 1.0, 'view factor to ground')

    @classmethod
    def from_dict(cls, data):
        """Initialize Outdoors BoundaryCondition from a dictionary.

        Args:
            data: A dictionary representation of the boundary condition.
        """
        assert data['type'] == 'Outdoors', 'Expected dictionary for Outdoors boundary ' \
            'condition. Got {}.'.format(data['type'])
        sun_exposure = True if 'sun_exposure' not in data else data['sun_exposure']
        wind_exposure = True if 'wind_exposure' not in data else data['wind_exposure']
        view_factor = autocalculate if 'view_factor' not in data or \
            data['view_factor'] == autocalculate.to_dict() else data['view_factor']
        return cls(sun_exposure, wind_exposure, view_factor)

    @property
    def sun_exposure(self):
        """Get a boolean noting whether the boundary is exposed to sun."""
        return self._sun_exposure

    @property
    def wind_exposure(self):
        """Get a boolean noting whether the boundary is exposed to wind."""
        return self._wind_exposure

    @property
    def view_factor(self):
        """Get the view factor to the ground as a number or 'autocalculate'."""
        return self._view_factor

    @property
    def sun_exposure_idf(self):
        """Get a text string for sun exposure, which is write-able into an IDF."""
        return 'NoSun' if not self.sun_exposure else 'SunExposed'

    @property
    def wind_exposure_idf(self):
        """Get a text string for wind exposure, which is write-able into an IDF."""
        return 'NoWind' if not self.wind_exposure else 'WindExposed'

    def to_dict(self, full=False):
        """Get the boundary condition as a dictionary.

        Args:
            full: Set to True to get the full dictionary which includes energy
                simulation specific keys such as sun_exposure, wind_exposure and
                view_factor. (Default: False).
        """
        bc_dict = {'type': self.name}
        if full:
            bc_dict['sun_exposure'] = self.sun_exposure
            bc_dict['wind_exposure'] = self.wind_exposure
            bc_dict['view_factor'] = autocalculate.to_dict() if \
                self.view_factor == autocalculate else self.view_factor
        return bc_dict

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return (self.sun_exposure, self.wind_exposure, self.view_factor)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Outdoors) and self.__key() == other.__key()


class Surface(_BoundaryCondition):
    """Boundary condition when an object is adjacent to another object."""

    __slots__ = ('_boundary_condition_objects',)

    def __init__(self, boundary_condition_objects, sub_face=False):
        """Initialize Surface boundary condition.

        Args:
            boundary_condition_objects: A list of up to 3 object identifiers that are
                adjacent to this one. The first object is always immediately
                adjacent and is of the same object type (Face, Aperture, Door). When
                this boundary condition is applied to a Face, the second object in the
                tuple will be the parent Room of the adjacent object. When the boundary
                condition is applied to a sub-face (Door or Aperture), the second object
                will be the parent Face of the adjacent sub-face and the third object
                 will be the parent Room of the adjacent sub-face.
            sub_face: Boolean to note whether this boundary condition is applied to a
                sub-face (an Aperture or a Door) instead of a Face. (Default: False).
        """
        if sub_face:
            self._boundary_condition_objects = tuple_with_length(
                boundary_condition_objects, 3, str,
                'boundary_condition_objects for Apertures or Doors')
        else:
            self._boundary_condition_objects = tuple_with_length(
                boundary_condition_objects, 2, str,
                'boundary_condition_objects for Faces')

    @classmethod
    def from_dict(cls, data, sub_face=False):
        """Initialize Surface BoundaryCondition from a dictionary.

        Args:
            data: A dictionary representation of the boundary condition.
            sub_face: Boolean to note whether this boundary condition is applied to a
                sub-face (an Aperture or a Door) instead of a Face. Default: False.
        """
        assert data['type'] == 'Surface', 'Expected dictionary for Surface boundary ' \
            'condition. Got {}.'.format(data['type'])
        return cls(data['boundary_condition_objects'], sub_face)

    @classmethod
    def from_other_object(cls, other_object, sub_face=False):
        """Initialize Surface boundary condition from an adjacent other object.

        Args:
            other_object: Another object (Face, Aperture, Door) of the same type
                that this boundary condition is assigned.  This other_object will be
                set as the adjacent object in this boundary condition.
            sub_face: Boolean to note whether this boundary condition is applied to a
                sub-face (an Aperture or a Door) instead of a Face. Default: False.
        """
        error_msg = 'Surface boundary conditions can only be assigned to objects' \
            ' with parent Rooms.'
        bc_objects = [other_object.identifier]
        if other_object.has_parent:
            bc_objects.append(other_object.parent.identifier)
            if sub_face:
                if other_object.parent.has_parent:
                    bc_objects.append(other_object.parent.parent.identifier)
                else:
                    raise AttributeError(error_msg)
        else:
            raise AttributeError(error_msg)
        return cls(bc_objects, sub_face)

    @property
    def boundary_condition_objects(self):
        """Get a tuple of up to 3 object identifiers that are adjacent to this one.

        The first object is always the one that is immediately adjacent and is of
        the same object type (Face, Aperture, Door).
        When this boundary condition is applied to a Face, the second object in the
        tuple will be the parent Room of the adjacent object.
        When the boundary condition is applied to a sub-face (Door or Aperture),
        the second object will be the parent Face of the sub-face and the third
        object will be the parent Room of the adjacent sub-face.
        """
        return self._boundary_condition_objects

    @property
    def boundary_condition_object(self):
        """Get the identifier of the object adjacent to this one."""
        return self._boundary_condition_objects[0]

    def to_dict(self):
        """Get the boundary condition as a dictionary.

        Args:
            full: Set to True to get the full dictionary which includes energy
                simulation specific keys such as sun_exposure, wind_exposure and
                view_factor. Default: False.
        """
        return {'type': self.name,
                'boundary_condition_objects': self.boundary_condition_objects}

    def __key(self):
        """A tuple based on the object properties, useful for hashing."""
        return self.boundary_condition_objects

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return isinstance(other, Surface) and self.__key() == other.__key()


class Ground(_BoundaryCondition):
    """Ground boundary condition.

    Args:
        data: A dictionary representation of the boundary condition.
    """
    __slots__ = ()

    @classmethod
    def from_dict(cls, data):
        """Initialize Ground BoundaryCondition from a dictionary."""
        assert data['type'] == 'Ground', 'Expected dictionary for Ground boundary ' \
            'condition. Got {}.'.format(data['type'])
        return cls()

    def __eq__(self, other):
        return isinstance(other, Ground)


class _BoundaryConditions(object):
    """Boundary conditions."""

    def __init__(self):
        self._outdoors = Outdoors()
        self._ground = Ground()
        self._bc_name_dict = None

    @property
    def outdoors(self):
        """Default outdoor boundary condition."""
        return self._outdoors

    @property
    def ground(self):
        """Default ground boundary condition."""
        return self._ground

    def surface(self, other_object, sub_face=False):
        """Get a Surface boundary condition.

        Args:
            other_object: The other object that is adjacent to the one that will
                bear this Surface boundary condition.
            sub_face: Boolean to note whether the boundary condition is for a
                sub-face (Aperture or Door) instead of a Face. (Default: False).
        """
        return Surface.from_other_object(other_object, sub_face)

    def by_name(self, bc_name):
        """Get a boundary condition object instance by its name.

        This method will correct for capitalization as well as the presence of
        spaces and underscores. Note that this method only works for boundary
        conditions with all of their inputs defaulted.

        Args:
            bc_name: A boundary condition name.
        """
        if self._bc_name_dict is None:
            self._build_bc_name_dict()
        try:
            return self._bc_name_dict[re.sub(r'[\s_]', '', bc_name.lower())]
        except KeyError:
            raise ValueError(
                '"{}" is not a valid boundary condition name.\nChoose from the '
                'following: {}'.format(bc_name, list(self._bc_name_dict.keys())))

    def _build_bc_name_dict(self):
        """Build a dictionary that can be used to lookup boundary conditions by name."""
        attr = [atr for atr in dir(self) if not atr.startswith('_')]
        clean_attr = [re.sub(r'[\s_]', '', atr.lower()) for atr in attr]
        self._bc_name_dict = {}
        for atr_name, atr in zip(clean_attr, attr):
            try:
                full_attr = getattr(self, '_' + atr)
                self._bc_name_dict[atr_name] = full_attr
            except AttributeError:
                pass  # callable method that has no static default object

    def __contains__(self, value):
        return isinstance(value, _BoundaryCondition)


boundary_conditions = _BoundaryConditions()


def get_bc_from_position(positions, ground_depth=0):
    """Return a boundary condition based on the relationship to a ground plane.

    Positions that are entirely at or below the ground_depth will get a Ground
    boundary condition. If there are any positions above the ground_depth, an
    Outdoors boundary condition will be returned.

    args:
        positions: A list of ladybug_geometry Point3D objects representing the
            vertices of an object.
        ground_depth: The Z value above which positions are considered Outdoors
            instead of Ground.

    Returns:
        Face type instance.
    """
    for position in positions:
        if position.z > ground_depth:
            return boundary_conditions.outdoors
    return boundary_conditions.ground



# coding: utf-8
"""Honeybee ShadeMesh."""
from __future__ import division
import math

from ladybug_geometry.geometry3d import Mesh3D, Face3D
from ladybug.color import Color

from ._base import _Base
from .typing import clean_string
from .properties import ShadeMeshProperties
import honeybee.writer.shademesh as writer


class ShadeMesh(_Base):
    """A single planar shade.

    Args:
        identifier: Text string for a unique Shade ID. Must be < 100 characters and
            not contain any spaces or special characters.
        geometry: A ladybug-geometry Mesh3D.
        is_detached: Boolean to note whether this object is detached from other
            geometry. Cases where this should be True include shade representing
            surrounding buildings or context. (Default: True).

    Properties:
        * identifier
        * display_name
        * is_detached
        * geometry
        * vertices
        * faces
        * center
        * area
        * min
        * max
        * type_color
        * bc_color
        * user_data
    """
    __slots__ = ('_geometry', '_is_detached')
    TYPE_COLORS = {
        False: Color(120, 75, 190),
        True: Color(80, 50, 128)
    }
    BC_COLOR = Color(120, 75, 190)

    def __init__(self, identifier, geometry, is_detached=True):
        """A single planar shade."""
        _Base.__init__(self, identifier)  # process the identifier

        # process the geometry and basic properties
        assert isinstance(geometry, Mesh3D), \
            'Expected ladybug_geometry Mesh3D. Got {}'.format(type(geometry))
        self._geometry = geometry
        self.is_detached = is_detached

        # initialize properties for extensions
        self._properties = ShadeMeshProperties(self)

    @classmethod
    def from_dict(cls, data):
        """Initialize an ShadeMesh from a dictionary.

        Args:
            data: A dictionary representation of an ShadeMesh object.
        """
        try:
            # check the type of dictionary
            assert data['type'] == 'ShadeMesh', 'Expected ShadeMesh dictionary. ' \
                'Got {}.'.format(data['type'])

            is_detached = data['is_detached'] if 'is_detached' in data else True
            shade = cls(
                data['identifier'], Mesh3D.from_dict(data['geometry']), is_detached)
            if 'display_name' in data and data['display_name'] is not None:
                shade.display_name = data['display_name']
            if 'user_data' in data and data['user_data'] is not None:
                shade.user_data = data['user_data']

            if data['properties']['type'] == 'ShadeMeshProperties':
                shade.properties._load_extension_attr_from_dict(data['properties'])
            return shade
        except Exception as e:
            cls._from_dict_error_message(data, e)

    @property
    def is_detached(self):
        """Get or set a boolean for whether this object is detached from other geometry.
        """
        return self._is_detached

    @is_detached.setter
    def is_detached(self, value):
        try:
            self._is_detached = bool(value)
        except TypeError:
            raise TypeError(
                'Expected boolean for ShadeMesh.is_detached. Got {}.'.format(value))

    @property
    def geometry(self):
        """Get a ladybug_geometry Mesh3D object representing the Shade."""
        return self._geometry

    @property
    def vertices(self):
        """Get a tuple of ladybug_geometry Point3D for the vertices of the mesh."""
        return self._geometry.vertices

    @property
    def faces(self):
        """Get a tuple of tuples for the faces of the mesh."""
        return self._geometry.faces

    @property
    def center(self):
        """Get a ladybug_geometry Point3D for the center of the shade.

        Note that this is the center of the bounding box around this geometry
        and not the area or volume centroid.
        """
        return self._geometry.center

    @property
    def area(self):
        """Get the surface area of the shade mesh."""
        return self._geometry.area

    @property
    def min(self):
        """Get a Point3D for the minimum of the bounding box around the object."""
        return self._geometry.min

    @property
    def max(self):
        """Get a Point3D for the maximum of the bounding box around the object."""
        return self._geometry.max

    @property
    def type_color(self):
        """Get a Color to be used in visualizations by type."""
        return self.TYPE_COLORS[self.is_detached]

    @property
    def bc_color(self):
        """Get a Color to be used in visualizations by boundary condition."""
        return self.BC_COLOR

    def add_prefix(self, prefix):
        """Change the identifier of this object by inserting a prefix.

        This is particularly useful in workflows where you duplicate and edit
        a starting object and then want to combine it with the original object
        into one Model (like making a model of repeated rooms) since all objects
        within a Model must have unique identifiers.

        Args:
            prefix: Text that will be inserted at the start of this object's identifier
                and display_name. It is recommended that this prefix be short to
                avoid maxing out the 100 allowable characters for honeybee identifiers.
        """
        self._identifier = clean_string('{}_{}'.format(prefix, self.identifier))
        self.display_name = '{}_{}'.format(prefix, self.display_name)
        self.properties.add_prefix(prefix)

    def move(self, moving_vec):
        """Move this Shade along a vector.

        Args:
            moving_vec: A ladybug_geometry Vector3D with the direction and distance
                to move the face.
        """
        self._geometry = self.geometry.move(moving_vec)
        self.properties.move(moving_vec)

    def rotate(self, axis, angle, origin):
        """Rotate this Shade by a certain angle around an axis and origin.

        Args:
            axis: A ladybug_geometry Vector3D axis representing the axis of rotation.
            angle: An angle for rotation in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        self._geometry = self.geometry.rotate(axis, math.radians(angle), origin)
        self.properties.rotate(axis, angle, origin)

    def rotate_xy(self, angle, origin):
        """Rotate this Shade counterclockwise in the world XY plane by a certain angle.

        Args:
            angle: An angle in degrees.
            origin: A ladybug_geometry Point3D for the origin around which the
                object will be rotated.
        """
        self._geometry = self.geometry.rotate_xy(math.radians(angle), origin)
        self.properties.rotate_xy(angle, origin)

    def reflect(self, plane):
        """Reflect this Shade across a plane.

        Args:
            plane: A ladybug_geometry Plane across which the object will
                be reflected.
        """
        self._geometry = self.geometry.reflect(plane.n, plane.o)
        self.properties.reflect(plane)

    def scale(self, factor, origin=None):
        """Scale this Shade by a factor from an origin point.

        Args:
            factor: A number representing how much the object should be scaled.
            origin: A ladybug_geometry Point3D representing the origin from which
                to scale. If None, it will be scaled from the World origin (0, 0, 0).
        """
        self._geometry = self.geometry.scale(factor, origin)
        self.properties.scale(factor, origin)

    def triangulate_and_remove_degenerate_faces(self, tolerance=0.01):
        """Triangulate non-planar faces in the mesh and remove all degenerate faces.

        This is helpful for certain geometry interfaces that require perfectly
        planar geometry without duplicate or colinear vertices.

        Args:
            tolerance: The minimum distance between a vertex and the boundary segments
                at which point the vertex is considered colinear. Default: 0.01,
                suitable for objects in meters.
        """
        new_faces, verts = [], self.geometry.vertices
        for shd in self.faces:
            shd_verts = [verts[v] for v in shd]
            shf = Face3D(shd_verts)
            if len(shd_verts) == 4 and not \
                    shf.check_planar(tolerance, raise_exception=False):
                shades = ((shd[0], shd[1], shd[2]), (shd[2], shd[3], shd[0]))
                for shade in shades:
                    shd_verts = [verts[v] for v in shade]
                    shade_face = Face3D(shd_verts)
                    try:
                        shade_face.remove_colinear_vertices(tolerance)
                    except AssertionError:
                        continue  # degenerate face to remove
                    new_faces.append(shade)
            else:
                try:
                    new_face = shf.remove_colinear_vertices(tolerance)
                except AssertionError:
                    continue  # degenerate face to remove
                if len(new_face.vertices) == len(shd):
                    new_faces.append(shd)
                else:  # quad face with duplicate or colinear verts
                    new_sh = tuple(shd[shd_verts.index(v)] for v in new_face.vertices)
                    new_faces.append(new_sh)
        self._geometry = Mesh3D(verts, new_faces)

    def is_geo_equivalent(self, shade_mesh, tolerance=0.01):
        """Get a boolean for whether this object is geometrically equivalent to another.

        The total number of vertices and the ordering of these vertices can be
        different but the geometries must share the same center point and be
        next to one another to within the tolerance.

        Args:
            shade_mesh: Another ShadeMesh for which geometric equivalency will be tested.
            tolerance: The minimum difference between the coordinate values of two
                vertices at which they can be considered geometrically equivalent.

        Returns:
            True if geometrically equivalent. False if not geometrically equivalent.
        """
        meta_1 = (self.display_name, self.is_detached)
        meta_2 = (shade_mesh.display_name, shade_mesh.is_detached)
        if meta_1 != meta_2:
            return False
        if len(self.geometry.vertices) != len(shade_mesh.geometry.vertices):
            return False
        if len(self.geometry.faces) != len(shade_mesh.geometry.faces):
            return False
        return all(pt.is_equivalent(o_pt, tolerance) for pt, o_pt in
                   zip(self.geometry.vertices, shade_mesh.geometry.vertices))

    def display_dict(self):
        """Get a list of DisplayMesh3D dictionaries for visualizing the object."""
        return [self._display_mesh(self.geometry, self.type_color)]

    @property
    def to(self):
        """ShadeMesh writer object.

        Use this method to access Writer class to write the shade in different formats.

        Usage:

        .. code-block:: python

            shade_mesh.to.idf(shade) -> idf string.
            shade_mesh.to.radiance(shade) -> Radiance string.
        """
        return writer

    def to_dict(self, abridged=False, included_prop=None):
        """Return Shade as a dictionary.

        Args:
            abridged: Boolean to note whether the extension properties of the
                object (ie. modifiers, transmittance schedule) should be included in
                detail (False) or just referenced by identifier (True). Default: False.
            included_prop: List of properties to filter keys that must be included in
                output dictionary. For example ['energy'] will include 'energy' key if
                available in properties to_dict. By default all the keys will be
                included. To exclude all the keys from extensions use an empty list.
        """
        base = {'type': 'ShadeMesh'}
        base['identifier'] = self.identifier
        base['display_name'] = self.display_name
        base['properties'] = self.properties.to_dict(abridged, included_prop)
        base['geometry'] = self._geometry.to_dict()
        if not self.is_detached:
            base['is_detached'] = self.is_detached
        if self.user_data is not None:
            base['user_data'] = self.user_data
        return base

    def to_shades(self):
        """Return a list of Honeybee Shade objects derived from this ShadeMesh.

        Note that the resulting Shades may be degenerate or non-planar so it
        may be useful to call triangulate_and_remove_degenerate_faces on this
        object before converting to Shades.
        """
        from honeybee.shade import Shade  # imported here to avoid circular import
        shades = []
        for i, shade_geo in enumerate(self.geometry.face_vertices):
            shade_id = '{}_{}'.format(self.identifier, i)
            shade = Shade(shade_id, Face3D(shade_geo), self.is_detached)
            if self._display_name is not None:
                shade.display_name = '{} {}'.format(self.display_name, i)
            shades.append(shade)
        return shades

    @staticmethod
    def _display_mesh(mesh3d, color):
        """Create a DisplayMesh3D dictionary from a Mesh3D and color."""
        return {
            'type': 'DisplayMesh3D',
            'geometry': mesh3d.to_dict(),
            'color': color.to_dict(),
            'display_mode': 'SurfaceWithEdges'
        }

    def __copy__(self):
        new_shade = ShadeMesh(self.identifier, self.geometry, self.is_detached)
        new_shade._display_name = self._display_name
        new_shade._user_data = None if self.user_data is None else self.user_data.copy()
        new_shade._properties._duplicate_extension_attr(self._properties)
        return new_shade

    def __repr__(self):
        return 'ShadeMesh: %s' % self.display_name
